@@ -2,15 +2,16 @@ use std::collections::HashMap;
 
 use super::prelude::*;
 
-use crate::env::{EnvMode, EnvStack};
+use crate::env::{EnvMode, EnvSetMode, EnvStack};
 use crate::exec::exec_subshell;
+use crate::parser::ParserEnvSetMode;
 use crate::wutil::fish_iswalnum;
 
 const VAR_NAME_PREFIX: &wstr = L!("_flag_");
 
 localizable_consts!(
     BUILTIN_ERR_INVALID_OPT_SPEC
-    "%ls: Invalid option spec '%ls' at char '%lc'\n"
+    "%s: Invalid option spec '%s' at char '%c'\n"
 );
 
 #[derive(Default)]
@@ -145,8 +146,8 @@ fn check_for_mutually_exclusive_flags(
                         if flag1 > flag2 {
                             std::mem::swap(&mut flag1, &mut flag2);
                         }
-                        streams.err.append(wgettext_fmt!(
-                            "%ls: %ls %ls: options cannot be used together\n",
+                        streams.err.append(&wgettext_fmt!(
+                            "%s: %s %s: options cannot be used together\n",
                             opts.name,
                             flag1,
                             flag2
@@ -167,8 +168,8 @@ fn parse_exclusive_args(opts: &mut ArgParseCmdOpts, streams: &mut IoStreams) ->
     for raw_xflags in &opts.raw_exclusive_flags {
         let xflags: Vec<_> = raw_xflags.split(',').collect();
         if xflags.len() < 2 {
-            streams.err.append(wgettext_fmt!(
-                "%ls: exclusive flag string '%ls' is not valid\n",
+            streams.err.append(&wgettext_fmt!(
+                "%s: exclusive flag string '%s' is not valid\n",
                 opts.name,
                 raw_xflags
             ));
@@ -185,8 +186,8 @@ fn parse_exclusive_args(opts: &mut ArgParseCmdOpts, streams: &mut IoStreams) ->
                 // It's a long flag we store as its short flag equivalent.
                 exclusive_set.push(*short_equiv);
             } else {
-                streams.err.append(wgettext_fmt!(
-                    "%ls: exclusive flag '%ls' is not valid\n",
+                streams.err.append(&wgettext_fmt!(
+                    "%s: exclusive flag '%s' is not valid\n",
                     opts.name,
                     flag
                 ));
@@ -195,7 +196,7 @@ fn parse_exclusive_args(opts: &mut ArgParseCmdOpts, streams: &mut IoStreams) ->
         }
 
         // Store the set of exclusive flags for use when parsing the supplied set of arguments.
-        opts.exclusive_flag_sets.push(exclusive_set.to_vec());
+        opts.exclusive_flag_sets.push(exclusive_set.clone());
     }
     Ok(SUCCESS)
 }
@@ -214,8 +215,8 @@ fn parse_flag_modifiers<'args>(
         && s.char_at(0) != '!'
         && s.char_at(0) != '&'
     {
-        streams.err.append(wgettext_fmt!(
-            "%ls: Implicit int short flag '%lc' does not allow modifiers like '%lc'\n",
+        streams.err.append(&wgettext_fmt!(
+            "%s: Implicit int short flag '%c' does not allow modifiers like '%c'\n",
             opts.name,
             opt_spec.short_flag,
             s.char_at(0)
@@ -249,7 +250,7 @@ fn parse_flag_modifiers<'args>(
 
     if s.char_at(0) == '!' {
         if opt_spec.arg_type == ArgType::NoArgument {
-            streams.err.append(wgettext_fmt!(
+            streams.err.append(&wgettext_fmt!(
                 BUILTIN_ERR_INVALID_OPT_SPEC,
                 opts.name,
                 option_spec,
@@ -261,7 +262,7 @@ fn parse_flag_modifiers<'args>(
         // Move cursor to the end so we don't expect a long flag.
         s = s.slice_from(s.char_count());
     } else if !s.is_empty() {
-        streams.err.append(wgettext_fmt!(
+        streams.err.append(&wgettext_fmt!(
             BUILTIN_ERR_INVALID_OPT_SPEC,
             opts.name,
             option_spec,
@@ -276,8 +277,8 @@ fn parse_flag_modifiers<'args>(
     }
 
     if opts.options.contains_key(&opt_spec.short_flag) {
-        streams.err.append(wgettext_fmt!(
-            "%ls: Short flag '%lc' already defined\n",
+        streams.err.append(&wgettext_fmt!(
+            "%s: Short flag '%c' already defined\n",
             opts.name,
             opt_spec.short_flag
         ));
@@ -285,7 +286,7 @@ fn parse_flag_modifiers<'args>(
     }
 
     *opt_spec_str = s;
-    return true;
+    true
 }
 
 /// Parse the text following the short flag letter.
@@ -308,8 +309,8 @@ fn parse_option_spec_sep<'args>(
             *counter += 1;
         }
         if opts.implicit_int_flag != '\0' {
-            streams.err.append(wgettext_fmt!(
-                "%ls: Implicit int flag '%lc' already defined\n",
+            streams.err.append(&wgettext_fmt!(
+                "%s: Implicit int flag '%c' already defined\n",
                 opts.name,
                 opts.implicit_int_flag
             ));
@@ -327,7 +328,7 @@ fn parse_option_spec_sep<'args>(
             opt_spec.short_flag_valid = false;
             i += 1;
             if i == s.char_count() {
-                streams.err.append(wgettext_fmt!(
+                streams.err.append(&wgettext_fmt!(
                     BUILTIN_ERR_INVALID_OPT_SPEC,
                     opts.name,
                     option_spec,
@@ -339,7 +340,7 @@ fn parse_option_spec_sep<'args>(
         '/' => {
             i += 1; // the struct is initialized assuming short_flag_valid should be true
             if i == s.char_count() {
-                streams.err.append(wgettext_fmt!(
+                streams.err.append(&wgettext_fmt!(
                     BUILTIN_ERR_INVALID_OPT_SPEC,
                     opts.name,
                     option_spec,
@@ -350,8 +351,8 @@ fn parse_option_spec_sep<'args>(
         }
         '#' => {
             if opts.implicit_int_flag != '\0' {
-                streams.err.append(wgettext_fmt!(
-                    "%ls: Implicit int flag '%lc' already defined\n",
+                streams.err.append(&wgettext_fmt!(
+                    "%s: Implicit int flag '%c' already defined\n",
                     opts.name,
                     opts.implicit_int_flag
                 ));
@@ -384,7 +385,7 @@ fn parse_option_spec_sep<'args>(
     }
 
     *opt_spec_str = s.slice_from(i);
-    return true;
+    true
 }
 
 fn parse_option_spec<'args>(
@@ -394,8 +395,8 @@ fn parse_option_spec<'args>(
     streams: &mut IoStreams,
 ) -> bool {
     if option_spec.is_empty() {
-        streams.err.append(wgettext_fmt!(
-            "%ls: An option spec must have at least a short or a long flag\n",
+        streams.err.append(&wgettext_fmt!(
+            "%s: An option spec must have at least a short or a long flag\n",
             opts.name
         ));
         return false;
@@ -404,8 +405,8 @@ fn parse_option_spec<'args>(
     let mut s = option_spec;
     if !fish_iswalnum(s.char_at(0)) && s.char_at(0) != '#' && !(s.char_at(0) == '/' && s.len() > 1)
     {
-        streams.err.append(wgettext_fmt!(
-            "%ls: Short flag '%lc' invalid, must be alphanum or '#'\n",
+        streams.err.append(&wgettext_fmt!(
+            "%s: Short flag '%c' invalid, must be alphanum or '#'\n",
             opts.name,
             s.char_at(0)
         ));
@@ -431,8 +432,8 @@ fn parse_option_spec<'args>(
         if long_flag_char_count > 0 {
             opt_spec.long_flag = s.slice_to(long_flag_char_count);
             if opts.long_to_short_flag.contains_key(opt_spec.long_flag) {
-                streams.err.append(wgettext_fmt!(
-                    "%ls: Long flag '%ls' already defined\n",
+                streams.err.append(&wgettext_fmt!(
+                    "%s: Long flag '%s' already defined\n",
                     opts.name,
                     opt_spec.long_flag
                 ));
@@ -457,7 +458,7 @@ fn parse_option_spec<'args>(
     // Record our option under its short flag.
     opts.options.insert(opt_spec.short_flag, opt_spec);
 
-    return true;
+    true
 }
 
 fn collect_option_specs<'args>(
@@ -467,7 +468,7 @@ fn collect_option_specs<'args>(
     args: &[&'args wstr],
     streams: &mut IoStreams,
 ) -> BuiltinResult {
-    let Some(&cmd) = args.get(0) else {
+    let Some(&cmd) = args.first() else {
         return Err(STATUS_INVALID_ARGS);
     };
 
@@ -480,7 +481,7 @@ fn collect_option_specs<'args>(
         if *optind == argc {
             streams
                 .err
-                .append(wgettext_fmt!("%ls: Missing -- separator\n", cmd));
+                .append(&wgettext_fmt!("%s: Missing -- separator\n", cmd));
             return Err(STATUS_INVALID_ARGS);
         }
 
@@ -502,11 +503,11 @@ fn collect_option_specs<'args>(
     if counter > counter_max {
         streams
             .err
-            .append(wgettext_fmt!("%ls: Too many long-only options\n", cmd));
+            .append(&wgettext_fmt!("%s: Too many long-only options\n", cmd));
         return Err(STATUS_INVALID_ARGS);
     }
 
-    return Ok(SUCCESS);
+    Ok(SUCCESS)
 }
 
 fn parse_cmd_opts<'args>(
@@ -517,7 +518,7 @@ fn parse_cmd_opts<'args>(
     parser: &Parser,
     streams: &mut IoStreams,
 ) -> BuiltinResult {
-    let Some(&cmd) = args.get(0) else {
+    let Some(&cmd) = args.first() else {
         return Err(STATUS_INVALID_ARGS);
     };
 
@@ -532,7 +533,7 @@ fn parse_cmd_opts<'args>(
             's' => opts.stop_nonopt = true,
             'i' | 'u' => {
                 if opts.unknown_handling != UnknownHandling::Error {
-                    streams.err.append(wgettext_fmt!(
+                    streams.err.append(&wgettext_fmt!(
                         BUILTIN_ERR_COMBO2_EXCLUSIVE,
                         cmd,
                         "--ignore-unknown",
@@ -556,8 +557,8 @@ fn parse_cmd_opts<'args>(
                 } else if kind == L!("none") {
                     ArgType::NoArgument
                 } else {
-                    streams.err.append(wgettext_fmt!(
-                        "%ls: Invalid --unknown-arguments value '%ls'\n",
+                    streams.err.append(&wgettext_fmt!(
+                        "%s: Invalid --unknown-arguments value '%s'\n",
                         cmd,
                         kind
                     ));
@@ -573,8 +574,8 @@ fn parse_cmd_opts<'args>(
                 opts.min_args = {
                     let x = fish_wcstol(w.woptarg.unwrap()).unwrap_or(-1);
                     if x < 0 {
-                        streams.err.append(wgettext_fmt!(
-                            "%ls: Invalid --min-args value '%ls'\n",
+                        streams.err.append(&wgettext_fmt!(
+                            "%s: Invalid --min-args value '%s'\n",
                             cmd,
                             w.woptarg.unwrap()
                         ));
@@ -587,8 +588,8 @@ fn parse_cmd_opts<'args>(
                 opts.max_args = {
                     let x = fish_wcstol(w.woptarg.unwrap()).unwrap_or(-1);
                     if x < 0 {
-                        streams.err.append(wgettext_fmt!(
-                            "%ls: Invalid --max-args value '%ls'\n",
+                        streams.err.append(&wgettext_fmt!(
+                            "%s: Invalid --max-args value '%s'\n",
                             cmd,
                             w.woptarg.unwrap()
                         ));
@@ -642,7 +643,7 @@ fn parse_cmd_opts<'args>(
         // The user didn't specify any option specs.
         streams
             .err
-            .append(wgettext_fmt!("%ls: Missing -- separator\n", cmd));
+            .append(&wgettext_fmt!("%s: Missing -- separator\n", cmd));
         return Err(STATUS_INVALID_ARGS);
     }
 
@@ -655,7 +656,7 @@ fn parse_cmd_opts<'args>(
     }
 
     *optind = w.wopt_index;
-    return collect_option_specs(opts, optind, argc, args, streams);
+    collect_option_specs(opts, optind, argc, args, streams)
 }
 
 fn populate_option_strings<'args>(
@@ -699,24 +700,31 @@ fn validate_arg<'opts>(
         return Ok(SUCCESS);
     }
 
-    let vars = parser.vars();
-    vars.push(true /* new_scope */);
+    parser.vars().push(true /* new_scope */);
 
-    let env_mode = EnvMode::LOCAL | EnvMode::EXPORT;
-    vars.set_one(L!("_argparse_cmd"), env_mode, opts_name.to_owned());
+    let local_exported_mode = ParserEnvSetMode::new(EnvMode::LOCAL | EnvMode::EXPORT);
+    parser.set_one(
+        L!("_argparse_cmd"),
+        local_exported_mode,
+        opts_name.to_owned(),
+    );
     let flag_name = WString::from(VAR_NAME_PREFIX) + "name";
     if is_long_flag {
-        vars.set_one(&flag_name, env_mode, opt_spec.long_flag.to_owned());
-    } else {
-        vars.set_one(
+        parser.set_one(
             &flag_name,
-            env_mode,
+            local_exported_mode,
+            opt_spec.long_flag.to_owned(),
+        );
+    } else {
+        parser.set_one(
+            &flag_name,
+            local_exported_mode,
             WString::from_chars(vec![opt_spec.short_flag]),
         );
     }
-    vars.set_one(
+    parser.set_one(
         &(WString::from(VAR_NAME_PREFIX) + "value"),
-        env_mode,
+        local_exported_mode,
         woptarg.to_owned(),
     );
 
@@ -730,10 +738,10 @@ fn validate_arg<'opts>(
     );
 
     for output in cmd_output {
-        streams.err.append(output);
+        streams.err.append(&output);
         streams.err.append_char('\n');
     }
-    vars.pop();
+    parser.vars().pop(parser.is_repainting());
     retval.map(|()| SUCCESS)
 }
 
@@ -947,7 +955,7 @@ fn argparse_parse_flags<'args>(
                         streams,
                     )?;
                 } else if opts.unknown_handling == UnknownHandling::Error {
-                    streams.err.append(wgettext_fmt!(
+                    streams.err.append(&wgettext_fmt!(
                         BUILTIN_ERR_UNKNOWN,
                         opts.name,
                         args_read[w.wopt_index - 1]
@@ -1002,7 +1010,7 @@ fn argparse_parse_flags<'args>(
                             Some(w.argv[w.wopt_index - 1])
                         } else {
                             // the option is at the end of argv, so it has no argument
-                            streams.err.append(wgettext_fmt!(
+                            streams.err.append(&wgettext_fmt!(
                                 BUILTIN_ERR_MISSING,
                                 opts.name,
                                 args_read[w.wopt_index - 1]
@@ -1018,7 +1026,7 @@ fn argparse_parse_flags<'args>(
                         && is_long_flag
                         && arg_contents.contains('=')
                     {
-                        streams.err.append(wgettext_fmt!(
+                        streams.err.append(&wgettext_fmt!(
                             BUILTIN_ERR_UNEXP_ARG,
                             opts.name,
                             args_read[w.wopt_index - 1]
@@ -1080,7 +1088,7 @@ fn argparse_parse_flags<'args>(
     opts.args_opts = w.argv_opts;
 
     *optind = w.wopt_index;
-    return Ok(SUCCESS);
+    Ok(SUCCESS)
 }
 
 // This function mimics the `next_opt()` usage found elsewhere in our other builtin commands.
@@ -1115,7 +1123,7 @@ fn check_min_max_args_constraints(
     let cmd = &opts.name;
 
     if opts.args.len() < opts.min_args {
-        streams.err.append(wgettext_fmt!(
+        streams.err.append(&wgettext_fmt!(
             BUILTIN_ERR_MIN_ARG_COUNT1,
             cmd,
             opts.min_args,
@@ -1125,7 +1133,7 @@ fn check_min_max_args_constraints(
     }
 
     if opts.max_args != usize::MAX && opts.args.len() > opts.max_args {
-        streams.err.append(wgettext_fmt!(
+        streams.err.append(&wgettext_fmt!(
             BUILTIN_ERR_MAX_ARG_COUNT1,
             cmd,
             opts.max_args,
@@ -1138,7 +1146,7 @@ fn check_min_max_args_constraints(
 }
 
 /// Put the result of parsing the supplied args into the caller environment as local vars.
-fn set_argparse_result_vars(vars: &EnvStack, opts: ArgParseCmdOpts) {
+fn set_argparse_result_vars(vars: &EnvStack, local_mode: EnvSetMode, opts: ArgParseCmdOpts) {
     for opt_spec in opts.options.values() {
         if opt_spec.num_seen == 0 {
             continue;
@@ -1147,7 +1155,7 @@ fn set_argparse_result_vars(vars: &EnvStack, opts: ArgParseCmdOpts) {
         if opt_spec.short_flag_valid {
             let mut var_name = WString::from(VAR_NAME_PREFIX);
             var_name.push(opt_spec.short_flag);
-            vars.set(&var_name, EnvMode::LOCAL, opt_spec.vals.clone());
+            vars.set(&var_name, local_mode, opt_spec.vals.clone());
         }
 
         if !opt_spec.long_flag.is_empty() {
@@ -1158,14 +1166,14 @@ fn set_argparse_result_vars(vars: &EnvStack, opts: ArgParseCmdOpts) {
                 .chars()
                 .map(|c| if fish_iswalnum(c) { c } else { '_' });
             let var_name_long: WString = VAR_NAME_PREFIX.chars().chain(long_flag).collect();
-            vars.set(&var_name_long, EnvMode::LOCAL, opt_spec.vals.clone());
+            vars.set(&var_name_long, local_mode, opt_spec.vals.clone());
         }
     }
 
     let args = opts.args.into_iter().map(|s| s.into_owned()).collect();
-    vars.set(L!("argv"), EnvMode::LOCAL, args);
+    vars.set(L!("argv"), local_mode, args);
     let args_opts = opts.args_opts.into_iter().map(|s| s.into_owned()).collect();
-    vars.set(L!("argv_opts"), EnvMode::LOCAL, args_opts);
+    vars.set(L!("argv_opts"), local_mode, args_opts);
 }
 
 /// The argparse builtin. This is explicitly not compatible with the BSD or GNU version of this
@@ -1176,7 +1184,7 @@ fn set_argparse_result_vars(vars: &EnvStack, opts: ArgParseCmdOpts) {
 /// version is a builtin it can directly set variables local to the current scope (e.g., a
 /// function). It doesn't need to write anything to stdout that then needs to be eval'd.
 pub fn argparse(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) -> BuiltinResult {
-    let Some(&cmd) = args.get(0) else {
+    let Some(&cmd) = args.first() else {
         return Err(STATUS_INVALID_ARGS);
     };
 
@@ -1213,7 +1221,11 @@ pub fn argparse(parser: &Parser, streams: &mut IoStreams, args: &mut [&wstr]) ->
 
     check_min_max_args_constraints(&opts, streams)?;
 
-    set_argparse_result_vars(parser.vars(), opts);
+    set_argparse_result_vars(
+        parser.vars(),
+        parser.convert_env_set_mode(ParserEnvSetMode::new(EnvMode::LOCAL)),
+        opts,
+    );
 
     Ok(SUCCESS)
 }
