@@ -7,34 +7,33 @@ use crate::ast::{
 use crate::builtins::shared::builtin_exists;
 use crate::color::Color;
 use crate::common::{
-    valid_var_name, valid_var_name_char, ASCII_MAX, EXPAND_RESERVED_BASE, EXPAND_RESERVED_END,
+    ASCII_MAX, EXPAND_RESERVED_BASE, EXPAND_RESERVED_END, valid_var_name, valid_var_name_char,
 };
 use crate::complete::complete_wrap_map;
 use crate::env::{EnvVar, Environment};
 use crate::expand::{
-    expand_one, expand_to_command_and_args, ExpandFlags, ExpandResultCode, PROCESS_EXPAND_SELF_STR,
+    ExpandFlags, ExpandResultCode, PROCESS_EXPAND_SELF_STR, expand_one, expand_to_command_and_args,
 };
 use crate::function;
-use crate::future_feature_flags::{feature_test, FeatureFlag};
+use crate::future_feature_flags::{FeatureFlag, feature_test};
 use crate::highlight::file_tester::FileTester;
-use crate::history::{all_paths_are_valid, HistoryItem};
+use crate::history::{HistoryItem, all_paths_are_valid};
 use crate::operation_context::OperationContext;
 use crate::parse_constants::{
     ParseKeyword, ParseTokenType, ParseTreeFlags, SourceRange, StatementDecoration,
 };
 use crate::parse_util::{
-    parse_util_locate_cmdsubst_range, parse_util_slice_length, MaybeParentheses,
+    MaybeParentheses, parse_util_locate_cmdsubst_range, parse_util_slice_length,
 };
 use crate::path::{path_as_implicit_cd, path_get_cdpath, path_get_path, paths_are_same_file};
 use crate::terminal::Outputter;
-use crate::text_face::{parse_text_face, TextFace, UnderlineStyle};
+use crate::text_face::{SpecifiedTextFace, TextFace, UnderlineStyle, parse_text_face};
 use crate::threads::assert_is_background_thread;
-use crate::tokenizer::{variable_assignment_equals_pos, PipeOrRedir};
-use crate::wchar::{wstr, WString, L};
-use crate::wchar_ext::WExt;
+use crate::tokenizer::{PipeOrRedir, variable_assignment_equals_pos};
 use crate::wcstringutil::string_prefixes_string;
-use std::collections::hash_map::Entry;
+use fish_wchar::{L, WExt, WString, wstr};
 use std::collections::HashMap;
+use std::collections::hash_map::Entry;
 
 use super::file_tester::IsFile;
 
@@ -76,6 +75,9 @@ pub fn colorize(text: &wstr, colors: &[HighlightSpec], vars: &dyn Environment) -
             outp.set_text_face(face);
             last_color = color;
         }
+        if i + 1 == text.char_count() && c == '\n' {
+            outp.set_text_face(TextFace::default());
+        }
         outp.writech(c);
     }
     outp.set_text_face(TextFace::default());
@@ -104,6 +106,22 @@ pub fn highlight_shell(
     *color = highlighter.highlight();
 }
 
+pub fn highlight_and_colorize(
+    text: &wstr,
+    ctx: &OperationContext<'_>,
+    vars: &dyn Environment,
+) -> Vec<u8> {
+    let mut colors = Vec::new();
+    highlight_shell(
+        text,
+        &mut colors,
+        ctx,
+        /*io_ok=*/ false,
+        /*cursor=*/ None,
+    );
+    colorize(text, &colors, vars)
+}
+
 /// highlight_color_resolver_t resolves highlight specs (like "a command") to actual RGB colors.
 /// It maintains a cache with no invalidation mechanism. The lifetime of these should typically be
 /// one screen redraw.
@@ -134,17 +152,25 @@ impl HighlightColorResolver {
             }
         }
     }
-    pub(crate) fn resolve_spec_uncached(
-        highlight: &HighlightSpec,
-        vars: &dyn Environment,
-    ) -> TextFace {
+    fn resolve_spec_uncached(highlight: &HighlightSpec, vars: &dyn Environment) -> TextFace {
         let resolve_role = |role| {
-            vars.get_unless_empty(get_highlight_var_name(role))
-                .or_else(|| vars.get_unless_empty(get_highlight_var_name(get_fallback(role))))
-                .or_else(|| vars.get_unless_empty(get_highlight_var_name(HighlightRole::normal)))
-                .as_ref()
-                .map(parse_text_face_for_highlight)
-                .unwrap_or_else(TextFace::default)
+            let mut roles: &[HighlightRole] = &[role, get_fallback(role), HighlightRole::normal];
+            // TODO(MSRV>=?) partition_dedup
+            for i in [2, 1] {
+                if roles[i - 1] == roles[i] {
+                    roles = &roles[..i];
+                }
+            }
+            for &role in roles {
+                if let Some(face) = vars
+                    .get_unless_empty(get_highlight_var_name(role))
+                    .as_ref()
+                    .and_then(parse_text_face_for_highlight)
+                {
+                    return face;
+                }
+            }
+            TextFace::default()
         };
         let mut face = resolve_role(highlight.foreground);
 
@@ -161,7 +187,8 @@ impl HighlightColorResolver {
         if highlight.valid_path {
             if let Some(valid_path_var) = vars.get(L!("fish_color_valid_path")) {
                 // Historical behavior is to not apply background.
-                let valid_path_face = parse_text_face_for_highlight(&valid_path_var);
+                let valid_path_face =
+                    parse_text_face_for_highlight(&valid_path_var).unwrap_or_default();
                 // Apply the foreground, except if it's normal. The intention here is likely
                 // to only override foreground if the valid path color has an explicit foreground.
                 if !valid_path_face.fg.is_normal() {
@@ -180,19 +207,21 @@ impl HighlightColorResolver {
 }
 
 /// Return the internal color code representing the specified color.
-pub(crate) fn parse_text_face_for_highlight(var: &EnvVar) -> TextFace {
+pub(crate) fn parse_text_face_for_highlight(var: &EnvVar) -> Option<TextFace> {
     let face = parse_text_face(var.as_list());
-    let default = TextFace::default();
-    let fg = face.fg.unwrap_or(default.fg);
-    let bg = face.bg.unwrap_or(default.bg);
-    let underline_color = face.underline_color.unwrap_or(default.underline_color);
-    let style = face.style;
-    TextFace {
-        fg,
-        bg,
-        underline_color,
-        style,
-    }
+    (face != SpecifiedTextFace::default()).then(|| {
+        let default = TextFace::default();
+        let fg = face.fg.unwrap_or(default.fg);
+        let bg = face.bg.unwrap_or(default.bg);
+        let underline_color = face.underline_color.unwrap_or(default.underline_color);
+        let style = face.style.unwrap_or_default();
+        TextFace {
+            fg,
+            bg,
+            underline_color,
+            style,
+        }
+    })
 }
 
 fn command_is_valid(
@@ -209,14 +238,14 @@ fn command_is_valid(
     let mut implicit_cd_ok = true;
     if matches!(
         decoration,
-        StatementDecoration::command | StatementDecoration::exec
+        StatementDecoration::Command | StatementDecoration::Exec
     ) {
         builtin_ok = false;
         function_ok = false;
         abbreviation_ok = false;
         command_ok = true;
         implicit_cd_ok = false;
-    } else if decoration == StatementDecoration::builtin {
+    } else if decoration == StatementDecoration::Builtin {
         builtin_ok = true;
         function_ok = false;
         abbreviation_ok = false;
@@ -253,7 +282,7 @@ fn command_is_valid(
     }
 
     // Return what we got.
-    return is_valid;
+    is_valid
 }
 
 fn has_expand_reserved(s: &wstr) -> bool {
@@ -279,13 +308,13 @@ fn autosuggest_parse_command(
 
     // Find the first statement.
     let job_list: &ast::JobList = ast.top();
-    let jc = job_list.get(0)?;
+    let jc = job_list.first()?;
     let first_statement = jc.job.statement.as_decorated_statement()?;
 
     if let Some(expanded_command) = statement_get_expanded_command(buff, first_statement, ctx) {
         let mut arg = WString::new();
         // Check if the first argument or redirection is, in fact, an argument.
-        if let Some(arg_or_redir) = first_statement.args_or_redirs.get(0) {
+        if let Some(arg_or_redir) = first_statement.args_or_redirs.first() {
             if arg_or_redir.is_argument() {
                 arg = arg_or_redir.argument().source(buff).to_owned();
             }
@@ -318,21 +347,20 @@ pub fn autosuggest_validate_from_history(
     };
 
     // We handle cd specially.
-    if is_veritable_cd(&parsed_command) && !cd_dir.is_empty() {
-        if expand_one(&mut cd_dir, ExpandFlags::FAIL_ON_CMDSUBST, ctx, None) {
-            if string_prefixes_string(&cd_dir, L!("--help"))
-                || string_prefixes_string(&cd_dir, L!("-h"))
-            {
-                // cd --help is always valid.
-                return true;
-            } else {
-                // Check the directory target, respecting CDPATH.
-                // Permit the autosuggestion if the path is valid and not our directory.
-                let path = path_get_cdpath(&cd_dir, working_directory, ctx.vars());
-                return path
-                    .map(|p| !paths_are_same_file(working_directory, &p))
-                    .unwrap_or(false);
-            }
+    if is_veritable_cd(&parsed_command)
+        && !cd_dir.is_empty()
+        && expand_one(&mut cd_dir, ExpandFlags::FAIL_ON_CMDSUBST, ctx, None)
+    {
+        if string_prefixes_string(&cd_dir, L!("--help"))
+            || string_prefixes_string(&cd_dir, L!("-h"))
+        {
+            // cd --help is always valid.
+            return true;
+        } else {
+            // Check the directory target, respecting CDPATH.
+            // Permit the autosuggestion if the path is valid and not our directory.
+            let path = path_get_cdpath(&cd_dir, working_directory, ctx.vars());
+            return path.is_some_and(|p| !paths_are_same_file(working_directory, &p));
         }
     }
 
@@ -554,7 +582,7 @@ fn color_string_internal(buffstr: &wstr, base_color: HighlightSpec, colors: &mut
                             in_pos -= 1;
                         }
                         '?' => {
-                            if !feature_test(FeatureFlag::qmark_noglob) {
+                            if !feature_test(FeatureFlag::QuestionMarkNoGlob) {
                                 colors[in_pos] = HighlightSpec::with_fg(HighlightRole::operat);
                             }
                         }
@@ -863,14 +891,14 @@ impl<'s> Highlighter<'s> {
     fn visit_token(&mut self, tok: &dyn Token) {
         let mut role = HighlightRole::normal;
         match tok.token_type() {
-            ParseTokenType::end | ParseTokenType::pipe | ParseTokenType::background => {
+            ParseTokenType::End | ParseTokenType::Pipe | ParseTokenType::Background => {
                 role = HighlightRole::statement_terminator
             }
-            ParseTokenType::left_brace | ParseTokenType::right_brace => {
+            ParseTokenType::LeftBrace | ParseTokenType::RightBrace => {
                 role = HighlightRole::keyword;
             }
-            ParseTokenType::andand | ParseTokenType::oror => role = HighlightRole::operat,
-            ParseTokenType::string => {
+            ParseTokenType::AndAnd | ParseTokenType::OrOr => role = HighlightRole::operat,
+            ParseTokenType::String => {
                 // Assume all strings are params. This handles e.g. the variables a for header or
                 // function header. Other strings (like arguments to commands) need more complex
                 // handling, which occurs in their respective overrides of visit().
@@ -939,24 +967,29 @@ impl<'s> Highlighter<'s> {
         }
         // No command substitution, so we can highlight the target file or fd. For example,
         // disallow redirections into a non-existent directory.
-        let target_is_valid = if !self.io_still_ok() {
+        let (role, file_exists) = if !self.io_still_ok() {
             // I/O is disallowed, so we don't have much hope of catching anything but gross
             // errors. Assume it's valid.
-            true
+            (HighlightRole::redirection, false)
         } else if contains_pending_variable(&self.pending_variables, &target) {
-            true
+            // Target uses a variable defined by the current commandline. Assume it's valid.
+            (HighlightRole::redirection, false)
         } else {
             // Validate the redirection target..
-            self.file_tester.test_redirection_target(&target, oper.mode)
-        };
-        self.color_node(
-            &redir.target,
-            HighlightSpec::with_fg(if target_is_valid {
-                HighlightRole::redirection
+            if let Ok(IsFile(file_exists)) =
+                self.file_tester.test_redirection_target(&target, oper.mode)
+            {
+                (HighlightRole::redirection, file_exists)
             } else {
-                HighlightRole::error
-            }),
-        );
+                (HighlightRole::error, false)
+            }
+        };
+        self.color_node(&redir.target, HighlightSpec::with_fg(role));
+        if file_exists {
+            for i in redir.target.source_range().as_usize() {
+                self.color_array[i].valid_path = true;
+            }
+        }
     }
 
     fn visit_variable_assignment(&mut self, varas: &VariableAssignment) {
@@ -1071,9 +1104,9 @@ impl<'s> Highlighter<'s> {
 fn has_cmdsub(src: &wstr) -> bool {
     let mut cursor = 0;
     match parse_util_locate_cmdsubst_range(src, &mut cursor, true, None, None) {
-        MaybeParentheses::Error => return false,
-        MaybeParentheses::None => return false,
-        MaybeParentheses::CommandSubstitution(_) => return true,
+        MaybeParentheses::Error => false,
+        MaybeParentheses::None => false,
+        MaybeParentheses::CommandSubstitution(_) => true,
     }
 }
 
@@ -1106,7 +1139,7 @@ impl<'s, 'a> NodeVisitor<'a> for Highlighter<'s> {
             return self.visit_keyword(keyword);
         }
         if let Some(token) = node.as_token() {
-            if token.token_type() == ParseTokenType::end {
+            if token.token_type() == ParseTokenType::End {
                 self.visit_semi_nl(node);
                 return;
             }
@@ -1211,28 +1244,12 @@ fn get_fallback(role: HighlightRole) -> HighlightRole {
     }
 }
 
-impl Default for HighlightRole {
-    fn default() -> Self {
-        Self::normal
-    }
-}
-
-impl Default for HighlightSpec {
-    fn default() -> Self {
-        Self {
-            foreground: Default::default(),
-            background: Default::default(),
-            valid_path: Default::default(),
-            force_underline: Default::default(),
-        }
-    }
-}
-
 /// Describes the role of a span of text.
-#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
+#[derive(Clone, Copy, Debug, Default, Eq, Hash, PartialEq)]
 #[repr(u8)]
 pub enum HighlightRole {
-    normal,  // normal text
+    #[default]
+    normal, // normal text
     error,   // error
     command, // command
     keyword,
@@ -1266,10 +1283,600 @@ pub enum HighlightRole {
 }
 
 /// Simple value type describing how a character should be highlighted.
-#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq)]
+#[derive(Clone, Copy, Debug, Default, Eq, Hash, PartialEq)]
 pub struct HighlightSpec {
     pub foreground: HighlightRole,
     pub background: HighlightRole,
     pub valid_path: bool,
     pub force_underline: bool,
 }
+
+#[cfg(test)]
+mod tests {
+    use super::{HighlightColorResolver, HighlightRole, HighlightSpec, highlight_shell};
+    use crate::common::ScopeGuard;
+    use crate::env::{EnvMode, EnvSetMode, Environment};
+    use crate::future_feature_flags::{self, FeatureFlag};
+    use crate::highlight::parse_text_face_for_highlight;
+    use crate::operation_context::{EXPANSION_LIMIT_BACKGROUND, OperationContext};
+    use crate::prelude::*;
+    use crate::tests::prelude::*;
+    use crate::text_face::UnderlineStyle;
+    use libc::PATH_MAX;
+
+    // Helper to return a string whose length greatly exceeds PATH_MAX.
+    fn get_overlong_path() -> String {
+        let path_max = usize::try_from(PATH_MAX).unwrap();
+        let mut longpath = String::with_capacity(path_max * 2 + 10);
+        while longpath.len() <= path_max * 2 {
+            longpath += "/overlong";
+        }
+        longpath
+    }
+
+    #[test]
+    #[serial]
+    fn test_highlighting() {
+        let _cleanup = test_init();
+        let parser = TestParser::new();
+        // Testing syntax highlighting
+        parser.pushd("test/fish_highlight_test/");
+        let _popd = ScopeGuard::new((), |_| parser.popd());
+        std::fs::create_dir_all("dir").unwrap();
+        std::fs::create_dir_all("cdpath-entry/dir-in-cdpath").unwrap();
+        std::fs::write("foo", []).unwrap();
+        std::fs::write("bar", []).unwrap();
+
+        // Here are the components of our source and the colors we expect those to be.
+        #[derive(Debug)]
+        struct HighlightComponent<'a> {
+            text: &'a str,
+            color: HighlightSpec,
+            nospace: bool,
+        }
+
+        macro_rules! component {
+            ( ( $text:expr, $color:expr) ) => {
+                HighlightComponent {
+                    text: $text,
+                    color: $color,
+                    nospace: false,
+                }
+            };
+            ( ( $text:literal, $color:expr, ns ) ) => {
+                HighlightComponent {
+                    text: $text,
+                    color: $color,
+                    nospace: true,
+                }
+            };
+        }
+
+        macro_rules! validate {
+            ( $($comp:tt),* $(,)? ) => {
+                let components = [
+                    $(
+                        component!($comp),
+                    )*
+                ];
+                let vars = parser.vars();
+                // Generate the text.
+                let mut text = WString::new();
+                let mut expected_colors = vec![];
+                for comp in &components {
+                    if !text.is_empty() && !comp.nospace {
+                        text.push(' ');
+                        expected_colors.push(HighlightSpec::new());
+                    }
+                    text.push_str(comp.text);
+                    expected_colors.resize(text.len(), comp.color);
+                }
+                assert_eq!(text.len(), expected_colors.len());
+
+                let mut colors = vec![];
+                highlight_shell(
+                    &text,
+                    &mut colors,
+                    &OperationContext::background(vars, EXPANSION_LIMIT_BACKGROUND),
+                    true, /* io_ok */
+                    Some(text.len()),
+                );
+                assert_eq!(colors.len(), expected_colors.len());
+
+                for (i, c) in text.chars().enumerate() {
+                    // Hackish space handling. We don't care about the colors in spaces.
+                    if c == ' ' {
+                        continue;
+                    }
+
+                    assert_eq!(colors[i], expected_colors[i], "Failed at position {i}, char {c}");
+                }
+            };
+        }
+
+        let mut param_valid_path = HighlightSpec::with_fg(HighlightRole::param);
+        param_valid_path.valid_path = true;
+
+        let mut redirection_valid_path = HighlightSpec::with_fg(HighlightRole::redirection);
+        redirection_valid_path.valid_path = true;
+
+        let saved_flag = future_feature_flags::test(FeatureFlag::AmpersandNoBgInToken);
+        future_feature_flags::set(FeatureFlag::AmpersandNoBgInToken, true);
+        let _restore_saved_flag = ScopeGuard::new((), |_| {
+            future_feature_flags::set(FeatureFlag::AmpersandNoBgInToken, saved_flag);
+        });
+
+        let fg = HighlightSpec::with_fg;
+
+        // Verify variables and wildcards in commands using /bin/cat.
+        let vars = parser.vars();
+        let local_mode = EnvSetMode::new_at_early_startup(EnvMode::LOCAL);
+        vars.set_one(L!("CDPATH"), local_mode, L!("./cdpath-entry").to_owned());
+
+        vars.set_one(L!("VARIABLE_IN_COMMAND"), local_mode, L!("a").to_owned());
+        vars.set_one(L!("VARIABLE_IN_COMMAND2"), local_mode, L!("at").to_owned());
+
+        let _cleanup = ScopeGuard::new((), |_| {
+            vars.remove(L!("VARIABLE_IN_COMMAND"), EnvSetMode::default());
+            vars.remove(L!("VARIABLE_IN_COMMAND2"), EnvSetMode::default());
+        });
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("./foo", param_valid_path),
+            ("&", fg(HighlightRole::statement_terminator)),
+        );
+
+        validate!(
+            ("command", fg(HighlightRole::keyword)),
+            ("echo", fg(HighlightRole::command)),
+            ("abc", fg(HighlightRole::param)),
+            ("foo", param_valid_path),
+            ("&", fg(HighlightRole::statement_terminator)),
+        );
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("foo&bar", fg(HighlightRole::param)),
+            ("foo", fg(HighlightRole::param), ns),
+            ("&", fg(HighlightRole::statement_terminator)),
+            ("echo", fg(HighlightRole::command)),
+            ("&>", fg(HighlightRole::redirection)),
+        );
+
+        validate!(
+            ("if command", fg(HighlightRole::keyword)),
+            ("ls", fg(HighlightRole::command)),
+            ("; ", fg(HighlightRole::statement_terminator)),
+            ("echo", fg(HighlightRole::command)),
+            ("abc", fg(HighlightRole::param)),
+            ("; ", fg(HighlightRole::statement_terminator)),
+            ("/bin/definitely_not_a_command", fg(HighlightRole::error)),
+            ("; ", fg(HighlightRole::statement_terminator)),
+            ("end", fg(HighlightRole::keyword)),
+        );
+
+        validate!(
+            ("if", fg(HighlightRole::keyword)),
+            ("true", fg(HighlightRole::command)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("else", fg(HighlightRole::keyword)),
+            ("true", fg(HighlightRole::command)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("end", fg(HighlightRole::keyword)),
+        );
+
+        // Verify that cd shows errors for non-directories.
+        validate!(
+            ("cd", fg(HighlightRole::command)),
+            ("dir", param_valid_path),
+        );
+
+        validate!(
+            ("cd", fg(HighlightRole::command)),
+            ("foo", fg(HighlightRole::error)),
+        );
+
+        validate!(
+            ("cd", fg(HighlightRole::command)),
+            ("--help", fg(HighlightRole::option)),
+            ("-h", fg(HighlightRole::option)),
+            ("definitely_not_a_directory", fg(HighlightRole::error)),
+        );
+
+        validate!(
+            ("cd", fg(HighlightRole::command)),
+            ("dir-in-cdpath", param_valid_path),
+        );
+
+        // Command substitutions.
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("param1", fg(HighlightRole::param)),
+            ("-l", fg(HighlightRole::option)),
+            ("--", fg(HighlightRole::option)),
+            ("-l", fg(HighlightRole::param)),
+            ("(", fg(HighlightRole::operat)),
+            ("ls", fg(HighlightRole::command)),
+            ("-l", fg(HighlightRole::option)),
+            ("--", fg(HighlightRole::option)),
+            ("-l", fg(HighlightRole::param)),
+            ("param2", fg(HighlightRole::param)),
+            (")", fg(HighlightRole::operat)),
+            ("|", fg(HighlightRole::statement_terminator)),
+            ("cat", fg(HighlightRole::command)),
+        );
+        validate!(
+            ("true", fg(HighlightRole::command)),
+            ("$(", fg(HighlightRole::operat)),
+            ("true", fg(HighlightRole::command)),
+            (")", fg(HighlightRole::operat)),
+        );
+        validate!(
+            ("true", fg(HighlightRole::command)),
+            ("\"before", fg(HighlightRole::quote)),
+            ("$(", fg(HighlightRole::operat)),
+            ("true", fg(HighlightRole::command)),
+            ("param1", fg(HighlightRole::param)),
+            (")", fg(HighlightRole::operat)),
+            ("after\"", fg(HighlightRole::quote)),
+            ("param2", fg(HighlightRole::param)),
+        );
+        validate!(
+            ("true", fg(HighlightRole::command)),
+            ("\"", fg(HighlightRole::error)),
+            ("unclosed quote", fg(HighlightRole::quote)),
+            ("$(", fg(HighlightRole::operat)),
+            ("true", fg(HighlightRole::command)),
+            (")", fg(HighlightRole::operat)),
+        );
+
+        // Redirections substitutions.
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("param1", fg(HighlightRole::param)),
+            // Input redirection.
+            ("<", fg(HighlightRole::redirection)),
+            ("/dev/null", redirection_valid_path),
+            // Output redirection to a valid fd.
+            ("1>&2", fg(HighlightRole::redirection)),
+            // Output redirection to an invalid fd.
+            ("2>&", fg(HighlightRole::redirection)),
+            ("LO", fg(HighlightRole::error)),
+            // Just a param, not a redirection.
+            ("test/blah", fg(HighlightRole::param)),
+            // Input redirection from directory.
+            ("<", fg(HighlightRole::redirection)),
+            ("test/", fg(HighlightRole::error)),
+            // Output redirection to an invalid path.
+            ("3>", fg(HighlightRole::redirection)),
+            ("/not/a/valid/path/nope", fg(HighlightRole::error)),
+            // Output redirection to directory.
+            ("3>", fg(HighlightRole::redirection)),
+            ("test/nope/", fg(HighlightRole::error)),
+            // Redirections to overflow fd.
+            ("99999999999999999999>&2", fg(HighlightRole::error)),
+            ("2>&", fg(HighlightRole::redirection)),
+            ("99999999999999999999", fg(HighlightRole::error)),
+            // Output redirection containing a command substitution.
+            ("4>", fg(HighlightRole::redirection)),
+            ("(", fg(HighlightRole::operat)),
+            ("echo", fg(HighlightRole::command)),
+            ("test/somewhere", fg(HighlightRole::param)),
+            (")", fg(HighlightRole::operat)),
+            // Just another param.
+            ("param2", fg(HighlightRole::param)),
+        );
+
+        validate!(
+            ("for", fg(HighlightRole::keyword)),
+            ("x", fg(HighlightRole::param)),
+            ("in", fg(HighlightRole::keyword)),
+            ("set-by-for-1", fg(HighlightRole::param)),
+            ("set-by-for-2", fg(HighlightRole::param)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("echo", fg(HighlightRole::command)),
+            (">", fg(HighlightRole::redirection)),
+            ("$x", fg(HighlightRole::redirection)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("end", fg(HighlightRole::keyword)),
+        );
+
+        validate!(
+            ("set", fg(HighlightRole::command)),
+            ("x", fg(HighlightRole::param)),
+            ("set-by-set", fg(HighlightRole::param)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("echo", fg(HighlightRole::command)),
+            (">", fg(HighlightRole::redirection)),
+            ("$x", fg(HighlightRole::redirection)),
+            ("2>", fg(HighlightRole::redirection)),
+            ("$totally_not_x", fg(HighlightRole::error)),
+            ("<", fg(HighlightRole::redirection)),
+            ("$x_but_its_an_impostor", fg(HighlightRole::error)),
+        );
+
+        validate!(
+            ("x", fg(HighlightRole::param), ns),
+            ("=", fg(HighlightRole::operat), ns),
+            ("set-by-variable-override", fg(HighlightRole::param), ns),
+            ("echo", fg(HighlightRole::command)),
+            (">", fg(HighlightRole::redirection)),
+            ("$x", fg(HighlightRole::redirection)),
+        );
+
+        validate!(
+            ("end", fg(HighlightRole::error)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("if", fg(HighlightRole::keyword)),
+            ("end", fg(HighlightRole::error)),
+        );
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("'", fg(HighlightRole::error)),
+            ("single_quote", fg(HighlightRole::quote)),
+            ("$stuff", fg(HighlightRole::quote)),
+        );
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("\"", fg(HighlightRole::error)),
+            ("double_quote", fg(HighlightRole::quote)),
+            ("$stuff", fg(HighlightRole::operat)),
+        );
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("$foo", fg(HighlightRole::operat)),
+            ("\"", fg(HighlightRole::quote)),
+            ("$bar", fg(HighlightRole::operat)),
+            ("\"", fg(HighlightRole::quote)),
+            ("$baz[", fg(HighlightRole::operat)),
+            ("1 2..3", fg(HighlightRole::param)),
+            ("]", fg(HighlightRole::operat)),
+        );
+
+        validate!(
+            ("for", fg(HighlightRole::keyword)),
+            ("i", fg(HighlightRole::param)),
+            ("in", fg(HighlightRole::keyword)),
+            ("1 2 3", fg(HighlightRole::param)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("end", fg(HighlightRole::keyword)),
+        );
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("$$foo[", fg(HighlightRole::operat)),
+            ("1", fg(HighlightRole::param)),
+            ("][", fg(HighlightRole::operat)),
+            ("2", fg(HighlightRole::param)),
+            ("]", fg(HighlightRole::operat)),
+            ("[3]", fg(HighlightRole::param)), // two dollar signs, so last one is not an expansion
+        );
+
+        validate!(
+            ("cat", fg(HighlightRole::command)),
+            ("/dev/null", param_valid_path),
+            ("|", fg(HighlightRole::statement_terminator)),
+            // This is bogus, but we used to use "less" here and that doesn't have to be installed.
+            ("cat", fg(HighlightRole::command)),
+            ("2>", fg(HighlightRole::redirection)),
+        );
+
+        // Highlight path-prefixes only at the cursor.
+        validate!(
+            ("cat", fg(HighlightRole::command)),
+            ("/dev/nu", fg(HighlightRole::param)),
+            ("/dev/nu", param_valid_path),
+        );
+
+        validate!(
+            ("if", fg(HighlightRole::keyword)),
+            ("true", fg(HighlightRole::command)),
+            ("&&", fg(HighlightRole::operat)),
+            ("false", fg(HighlightRole::command)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("or", fg(HighlightRole::operat)),
+            ("false", fg(HighlightRole::command)),
+            ("||", fg(HighlightRole::operat)),
+            ("true", fg(HighlightRole::command)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("and", fg(HighlightRole::operat)),
+            ("not", fg(HighlightRole::operat)),
+            ("!", fg(HighlightRole::operat)),
+            ("true", fg(HighlightRole::command)),
+            (";", fg(HighlightRole::statement_terminator)),
+            ("end", fg(HighlightRole::keyword)),
+        );
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("%self", fg(HighlightRole::operat)),
+            ("not%self", fg(HighlightRole::param)),
+            ("self%not", fg(HighlightRole::param)),
+        );
+
+        validate!(
+            ("false", fg(HighlightRole::command)),
+            ("&|", fg(HighlightRole::statement_terminator)),
+            ("true", fg(HighlightRole::command)),
+        );
+
+        validate!(
+            ("HOME", fg(HighlightRole::param)),
+            ("=", fg(HighlightRole::operat), ns),
+            (".", fg(HighlightRole::param), ns),
+            ("VAR1", fg(HighlightRole::param)),
+            ("=", fg(HighlightRole::operat), ns),
+            ("VAL1", fg(HighlightRole::param), ns),
+            ("VAR", fg(HighlightRole::param)),
+            ("=", fg(HighlightRole::operat), ns),
+            ("false", fg(HighlightRole::command)),
+            ("|&", fg(HighlightRole::error)),
+            ("true", fg(HighlightRole::command)),
+            ("stuff", fg(HighlightRole::param)),
+        );
+
+        validate!(
+            ("echo", fg(HighlightRole::command)), // (
+            (")", fg(HighlightRole::error)),
+        );
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("stuff", fg(HighlightRole::param)),
+            ("# comment", fg(HighlightRole::comment)),
+        );
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("--", fg(HighlightRole::option)),
+            ("-s", fg(HighlightRole::param)),
+        );
+
+        // Overlong paths don't crash (#7837).
+        let overlong = get_overlong_path();
+        validate!(
+            ("touch", fg(HighlightRole::command)),
+            (&overlong, fg(HighlightRole::param)),
+        );
+
+        validate!(
+            ("a", fg(HighlightRole::param)),
+            ("=", fg(HighlightRole::operat), ns),
+        );
+
+        // Highlighting works across escaped line breaks (#8444).
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("$FISH_\\\n", fg(HighlightRole::operat)),
+            ("VERSION", fg(HighlightRole::operat), ns),
+        );
+
+        validate!(
+            ("/bin/ca", fg(HighlightRole::command), ns),
+            ("*", fg(HighlightRole::operat), ns)
+        );
+
+        validate!(
+            ("/bin/c", fg(HighlightRole::command), ns),
+            ("*", fg(HighlightRole::operat), ns)
+        );
+
+        validate!(
+            ("/bin/c", fg(HighlightRole::command), ns),
+            ("{$VARIABLE_IN_COMMAND}", fg(HighlightRole::operat), ns),
+            ("*", fg(HighlightRole::operat), ns)
+        );
+
+        validate!(
+            ("/bin/c", fg(HighlightRole::command), ns),
+            ("$VARIABLE_IN_COMMAND2", fg(HighlightRole::operat), ns)
+        );
+
+        validate!(("$EMPTY_VARIABLE", fg(HighlightRole::error)));
+        validate!(("\"$EMPTY_VARIABLE\"", fg(HighlightRole::error)));
+
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("\\UFDFD", fg(HighlightRole::escape)),
+        );
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("\\U10FFFF", fg(HighlightRole::escape)),
+        );
+        validate!(
+            ("echo", fg(HighlightRole::command)),
+            ("\\U110000", fg(HighlightRole::error)),
+        );
+
+        validate!(
+            (">", fg(HighlightRole::error)),
+            ("echo", fg(HighlightRole::error)),
+        );
+    }
+
+    /// Tests that trailing spaces after a command don't inherit the underline formatting of the
+    /// command.
+    #[test]
+    #[serial]
+    #[allow(clippy::needless_range_loop)]
+    fn test_trailing_spaces_after_command() {
+        let _cleanup = test_init();
+        let parser = TestParser::new();
+        let vars = parser.vars();
+
+        // First, set up fish_color_command to include underline
+        vars.set_one(
+            L!("fish_color_command"),
+            EnvSetMode::new_at_early_startup(EnvMode::LOCAL),
+            L!("--underline").to_owned(),
+        );
+
+        // Prepare a command with trailing spaces for highlighting
+        let text = L!("echo   ").to_owned(); // Command 'echo' followed by 3 spaces
+        let mut colors = vec![];
+        highlight_shell(
+            &text,
+            &mut colors,
+            &OperationContext::background(vars, EXPANSION_LIMIT_BACKGROUND),
+            true, /* io_ok */
+            Some(text.len()),
+        );
+
+        // Verify we have the right number of colors
+        assert_eq!(colors.len(), text.len());
+
+        // Create a resolver to check actual RGB colors with their attributes
+        let mut resolver = HighlightColorResolver::new();
+
+        // Check that 'echo' is underlined
+        for i in 0..4 {
+            let face = resolver.resolve_spec(&colors[i], vars);
+            assert_eq!(
+                face.style.underline_style(),
+                Some(UnderlineStyle::Single),
+                "Character at position {} of 'echo' should be underlined",
+                i
+            );
+        }
+
+        // Check that trailing spaces are NOT underlined
+        for i in 4..text.len() {
+            let face = resolver.resolve_spec(&colors[i], vars);
+            assert_eq!(
+                face.style.underline_style(),
+                None,
+                "Trailing space at position {} should NOT be underlined",
+                i
+            );
+        }
+    }
+
+    #[test]
+    #[serial]
+    fn test_resolve_role() {
+        let _cleanup = test_init();
+        let parser = TestParser::new();
+        let vars = parser.vars();
+
+        let set = |var: &wstr, value: Vec<WString>| {
+            vars.set(var, EnvSetMode::new(EnvMode::LOCAL, false), value);
+        };
+        set(L!("fish_color_normal"), vec![L!("normal").into()]);
+        set(
+            L!("fish_color_command"),
+            vec![L!("red").into(), L!("--bold").into()],
+        );
+        set(L!("fish_color_keyword"), vec![L!("--theme=default").into()]);
+
+        let keyword_spec = HighlightSpec::with_both(HighlightRole::keyword);
+        let face = HighlightColorResolver::resolve_spec_uncached(&keyword_spec, vars);
+
+        let command_face =
+            parse_text_face_for_highlight(&vars.get(L!("fish_color_command")).unwrap()).unwrap();
+        assert_eq!(face, command_face);
+    }
+}
